@@ -17,96 +17,120 @@ See the License for the specific language governing permissions and limitations
 
 """
 
-# Package ff3 implements the FF3 format-preserving encryption algorithm/scheme
+# Package ff3 implements the FF3-1 format-preserving encryption algorithm/scheme
 
 import logging
 import math
 from Crypto.Cipher import AES
-# from hexdump import hexdump
 import string
 
-DOMAIN_MIN =  1000000  # 1M is currently recommended in FF3-1
-NUM_ROUNDS =   8
-BLOCK_SIZE =   16  # aes.BlockSize
-TWEAK_LEN =    8 # TODO: change to 7 bytes when 56-bit test vectors for FF3-1 become available
-HALF_TWEAK_LEN = TWEAK_LEN // 2
-MAX_RADIX =    36  # python int supports radix 2..36
+# The recommendation in Draft SP 800-38G was strengthened to a requirement in Draft SP 800-38G Revision 1:
+# the minimum domain size for FF1 and FF3-1 is one million.
 
-def reverse_string(aString):
-    "func defined for clarity"
-    return aString[::-1]
+NUM_ROUNDS = 8
+BLOCK_SIZE = 16  # aes.BlockSize
+TWEAK_LEN = 8  # Original FF3 tweak length
+TWEAK_LEN_NEW = 7  # FF3-1 tweak length
+HALF_TWEAK_LEN = TWEAK_LEN // 2
+
+
+def reverse_string(txt):
+    """func defined for clarity"""
+    return txt[::-1]
 
 
 """
-FF3 can encode a string within a range of minLen..maxLen. The spec uses an alternating Feistel
+FF3 encodes a string within a range of minLen..maxLen. The spec uses an alternating Feistel
 with the following parameters:
     128 bit key length
     Cipher Block Chain (CBC-MAC) round function
-    64-bit tweak
+    64-bit (FF3) or 56-bit (FF3-1)tweak
     eight (8) rounds
     Modulo addition
 
-An encoded string representation of x in the given base. Base must be between 2 and 36, inclusive. The result
-uses the uses the lower-case letters 'a' to 'z' for digit values 10 to 35.  Currently unimplemented, the
-upper-case letters 'A' to 'Z' would represent digit values 36 to 61.
+An encoded string representation of x is in the given integer base, which must be at least 2. The 
+result uses the lower-case letters 'a' to 'z' for digit values 10 to 35 and upper-case letters 'A' to 'Z' for 
+digit values 36 to 61.
 
-FF3Cipher initializes a new FF3 Cipher object for encryption or decryption with radix, key and tweak parameters.
+Instead of specifying the base, an alphabet may be specified as a string of unique characters.
+For bases larger than 62, an explicit alphabet is mandatory.
 
-AES ECB has a block size of 128 bits (i.e 16 bytes). It can only process data in blocks of this size.  Also, ECB 
-is encrypt only, a second encryption decrypts the text.
+FF3Cipher initializes a new FF3 Cipher object for encryption or decryption with key, tweak and radix parameters. The
+default radix is 10, supporting encryption of decimal numbers.
 
+AES ECB is used as the cipher round value for XORing. ECB has a block size of 128 bits (i.e 16 bytes) and is 
+padded with zeros for blocks smaller than this size. ECB is used only in encrypt mode to generate this XOR value. 
+A Feistel decryption uses the same ECB encrypt value to decrypt the text. XOR is trivially invertible when you 
+know two of the arguments.
 """
 
 
 class FF3Cipher:
-    """Class FF3Cipher implements the FF3 format-preserving encryption algorithm"""
-    def __init__(self, key, tweak, radix=10, ):
+    """Class FF3Cipher implements the FF3 format-preserving encryption algorithm.
 
-        self.key = bytes.fromhex(key)
+    If a value of radix between 2 and 62 is specified, then that many characters
+    from the base 62 alphabet (digits + lowercase + uppercase latin) are used.
+    """
+    DOMAIN_MIN = 1_000_000  # 1M required in FF3-1
+    BASE62 = string.digits + string.ascii_lowercase + string.ascii_uppercase
+    BASE62_LEN = len(BASE62)
+    RADIX_MAX = 256  # Support 8-bit alphabets for now, requires test cases for larger values
+
+    def __init__(self, key, tweak, radix=10, ):
+        keybytes = bytes.fromhex(key)
         self.tweak = tweak
         self.radix = radix
+        if radix <= FF3Cipher.BASE62_LEN:
+            self.alphabet = FF3Cipher.BASE62[0:radix]
+        else:
+            self.alphabet = None
 
         # Calculate range of supported message lengths [minLen..maxLen]
-        # per original spec, radix^minLength >= 100.
-        self.minLen = math.ceil(math.log(DOMAIN_MIN) / math.log(radix))
+        # per revised spec, radix^minLength >= 1,000,000.
+        self.minLen = math.ceil(math.log(FF3Cipher.DOMAIN_MIN) / math.log(radix))
 
         # We simplify the specs log[radix](2^96) to 96/log2(radix) using the log base change rule
         self.maxLen = 2 * math.floor(96/math.log2(radix))
 
-        keyLen = len(self.key)
+        klen = len(keybytes)
 
         # Check if the key is 128, 192, or 256 bits = 16, 24, or 32 bytes
-        if keyLen not in (16, 24, 32):
-            raise ValueError(f'key length is {keyLen} but must be 128, 192, or 256 bits')
+        if klen not in (16, 24, 32):
+            raise ValueError(f'key length is {klen} but must be 128, 192, or 256 bits')
 
-        # While FF3 allows radices in [2, 2^16], there is a practical limit to 36 (alphanumeric)
-        # because python int only supports up to base 36.
-        if (radix < 2) or (radix > MAX_RADIX):
-            raise ValueError("radix must be between 2 and 36, inclusive")
+        # While FF3 allows radices in [2, 2^16], commonly useful range is 2..62
+        if (radix < 2) or (radix > FF3Cipher.RADIX_MAX):
+            raise ValueError("radix must be between 2 and 62, inclusive")
 
         # Make sure 2 <= minLength <= maxLength
-        if ((self.minLen < 2) or (self.maxLen < self.minLen)):
+        if (self.minLen < 2) or (self.maxLen < self.minLen):
             raise ValueError("minLen or maxLen invalid, adjust your radix")
 
         # AES block cipher in ECB mode with the block size derived based on the length of the key
         # Always use the reversed key since Encrypt and Decrypt call ciph expecting that
 
-        self.aesBlock = AES.new(reverse_string(self.key), AES.MODE_ECB)
+        self.aesCipher = AES.new(reverse_string(keybytes), AES.MODE_ECB)
+
+    # factory method to create a FF3Cipher object with a custom alphabet
+    @staticmethod
+    def withCustomAlphabet(key, tweak, alphabet):
+        c = FF3Cipher(key, tweak, len(alphabet))
+        c.alphabet = alphabet
+        return c
 
     def encrypt(self, plaintext):
         """Encrypts the plaintext string and returns a ciphertext of the same length and format"""
         return self.encrypt_with_tweak(plaintext, self.tweak)
 
     """
-    Fiestel structure
+    Feistel structure
 
             u length |  v length
-
             A block  |  B block
 
-                C = modulo function
+                C <- modulo function
 
-            B' <- C    | A' <- B
+            B' <- C  |  A' <- B
 
 
     Steps:
@@ -116,7 +140,7 @@ class FF3Cipher:
     Let A = X[1..u]
     Let B = X[u+1,n]
     Let T(L) = T[0..31] and T(R) = T[32..63]
-    for i <- 0 to 6 do
+    for i <- 0..7 do
         If is even, let m = u and W = T(R) Else let m = v and W = T(L)
         Let P = REV([NUM<radix>(Rev(B))]^12 || W ⊗ REV(i^4)
         Let Y = CIPH(P)
@@ -148,14 +172,11 @@ class FF3Cipher:
         if (n < self.minLen) or (n > self.maxLen):
             raise ValueError(f"message length {n} is not within min {self.minLen} and max {self.maxLen} bounds")
 
-        # Make sure the given the length of tweak in bits is 64
-        if len(tweakBytes) != TWEAK_LEN:
-            raise ValueError(f"tweak length {len(tweakBytes)} invalid: tweak must be 8 bytes, or 64 bits")
+        # Make sure the given the length of tweak in bits is 56 or 64
+        if len(tweakBytes) not in [TWEAK_LEN, TWEAK_LEN_NEW]:
+            raise ValueError(f"tweak length {len(tweakBytes)} invalid: tweak must be 56 or 64 bits")
 
-        # Check if the plaintext message is formatted in the current radix
-        x = int(plaintext, self.radix)
-        if x == 0:
-            raise ValueError("plaintext string is not within base/radix {self.radix}")
+        # Todo: Check message is in current radix
 
         # Calculate split point
         u = math.ceil(n / 2)
@@ -165,16 +186,13 @@ class FF3Cipher:
         A = plaintext[:u]
         B = plaintext[u:]
 
-        # Split the tweak
+        if len(tweakBytes) == TWEAK_LEN_NEW:
+            # FF3-1
+            tweakBytes = calculate_tweak64_ff3_1(tweakBytes)
+
         Tl = tweakBytes[:HALF_TWEAK_LEN]
         Tr = tweakBytes[HALF_TWEAK_LEN:]
-
-        logging.debug(f"Tweak:{tweak}")
-        logging.debug(tweakBytes)
-        # hexdump(tweakBytes)
-
-        # P is always 16 bytes
-        P = bytearray(BLOCK_SIZE)
+        logging.debug(f"Tweak: {tweak}, tweakBytes:{tweakBytes.hex()}")
 
         # Pre-calculate the modulus since it's only one of 2 values,
         # depending on whether i is even or odd
@@ -184,6 +202,9 @@ class FF3Cipher:
         logging.debug(f"modU: {modU} modV: {modV}")
 
         # Main Feistel Round, 8 times
+        #
+        # AES ECB requires the number of bits in the plaintext to be a multiple of
+        # the block size. Thus, we pad the input to 16 bytes
 
         for i in range(NUM_ROUNDS):
             # logging.debug(f"-------- Round {i}")
@@ -195,47 +216,19 @@ class FF3Cipher:
                 m = v
                 W = Tl
 
-            # Calculate P by XORing W, i into the first 4 bytes of P
-            # i only requires 1 byte, rest are 0 padding bytes
-            # Anything XOR 0 is itself, so only need to XOR the last byte
-
-            P[0] = W[0]
-            P[1] = W[1]
-            P[2] = W[2]
-            P[3] = W[3] ^ int(i)
-
-            # The remaining 12 bytes of P are for rev(B) with padding
-
-            numB = reverse_string(B)
-            numBBytes = int(numB, self.radix).to_bytes(12, "big")
-
-            # logging.debug(f"B: {B} numB: {numB} numBBytes: {numBBytes.hex()}")
-
-            P[BLOCK_SIZE-len(numBBytes):] = numBBytes
-
-            # print("P:    ", end='')
-            # hexdump(P)
-
-            # Calculate S by operating on P in place
+            # P is fixed-length 16 bytes
+            P = calculate_p(i, self.alphabet, W, B)
             revP = reverse_string(P)
 
-            # print("revP: ", end='')
-            # hexdump(revP)
+            S = self.aesCipher.encrypt(bytes(revP))
 
-            # P is fixed-length 16 bytes
-            revP = self.aesBlock.encrypt(bytes(revP))
-
-            S = reverse_string(revP)
-            # print("S:    ", end='')
-            # hexdump(S)
+            S = reverse_string(S)
+            # logging.debug("S:    ", S.hex())
 
             y = int.from_bytes(S, byteorder='big')
 
             # Calculate c
-            c = int(reverse_string(A), self.radix)
-
-            if c == 0:
-                raise ValueError(f"string {A} is not within base/radix")
+            c = decode_int_r(A,  self.alphabet)
 
             c = c + y
 
@@ -245,7 +238,7 @@ class FF3Cipher:
                 c = c % modV
 
             # logging.debug(f"m: {m} A: {A} c: {c} y: {y}")
-            C = base_conv_r(c, self.radix, int(m))
+            C = encode_int_r(c, self.alphabet, int(m))
 
             # Final steps
             A = B
@@ -275,33 +268,27 @@ class FF3Cipher:
         if (n < self.minLen) or (n > self.maxLen):
             raise ValueError(f"message length {n} is not within min {self.minLen} and max {self.maxLen} bounds")
 
-        # Make sure the given the length of tweak in bits is 64
-        if len(tweakBytes) != TWEAK_LEN:
+        # Make sure the given the length of tweak in bits is 56 or 64
+        if len(tweakBytes) not in [TWEAK_LEN, TWEAK_LEN_NEW]:
             raise ValueError(f"tweak length {len(tweakBytes)} invalid: tweak must be 8 bytes, or 64 bits")
 
-        # Check if the ciphertext message is formatted in the current radix
-        x = int(ciphertext, self.radix)
-        if x == 0:
-            raise ValueError("ciphertext string is not within base/radix {self.radix}")
+        # Todo: Check message is in current radix
 
         # Calculate split point
-        u = math.ceil((n) / 2)
+        u = math.ceil(n/2)
         v = n - u
 
         # Split the message
         A = ciphertext[:u]
         B = ciphertext[u:]
 
-        # Split the tweak
+        if len(tweakBytes) == TWEAK_LEN_NEW:
+            # FF3-1
+            tweakBytes = calculate_tweak64_ff3_1(tweakBytes)
+
         Tl = tweakBytes[:HALF_TWEAK_LEN]
         Tr = tweakBytes[HALF_TWEAK_LEN:]
-
-        logging.debug(f"Tweak: {tweak}")
-        logging.debug(tweakBytes)
-        # hexdump(tweakBytes)
-
-        # P is always 16 bytes
-        P = bytearray(BLOCK_SIZE)
+        logging.debug(f"Tweak: {tweak}, tweakBytes:{tweakBytes.hex()}")
 
         # Pre-calculate the modulus since it's only one of 2 values,
         # depending on whether i is even or odd
@@ -323,48 +310,19 @@ class FF3Cipher:
                 m = v
                 W = Tl
 
-            # Calculate P by XORing W, i into the first 4 bytes of P
-            # i only requires 1 byte, rest are 0 padding bytes
-            # Anything XOR 0 is itself, so only need to XOR the last byte
-
-            P[0] = W[0]
-            P[1] = W[1]
-            P[2] = W[2]
-            P[3] = W[3] ^ int(i)
-
-            # The remaining 12 bytes of P are for rev(A) with padding
-
-            numA = reverse_string(A)
-            numABytes = int(numA, self.radix).to_bytes(12, "big")
-
-            # logging.debug(f"A: {A} numA: {numA} numABytes: {numABytes.hex()}")
-
-            P[BLOCK_SIZE-len(numABytes):] = numABytes
-
-            # print("P:    ", end='')
-            # hexdump(P)
-
-            # Calculate S by operating on P in place
+            # P is fixed-length 16 bytes
+            P = calculate_p(i, self.alphabet, W, A)
             revP = reverse_string(P)
 
-            # print("revP: ", end='')
-            # hexdump(revP)
+            S = self.aesCipher.encrypt(bytes(revP))
+            S = reverse_string(S)
 
-            # P is fixed-length 16 bytes
-
-            revP = self.aesBlock.encrypt(bytes(revP))
-
-            S = reverse_string(revP)
-            # print("S:    ", end='')
-            # hexdump(S)
+            # logging.debug("S:    ", S.hex())
 
             y = int.from_bytes(S, byteorder='big')
 
             # Calculate c
-            c = int(reverse_string(B), self.radix)
-
-            if c == 0:
-                raise ValueError("string A is not within base/radix")
+            c = decode_int_r(B, self.alphabet)
 
             c = c - y
 
@@ -373,8 +331,8 @@ class FF3Cipher:
             else:
                 c = c % modV
 
-            # logging.debug(f"m: {m} A: {A} c: {c} y: {y}")
-            C = base_conv_r(c, self.radix, int(m))
+            # logging.debug(f"m: {m} B: {B} c: {c} y: {y}")
+            C = encode_int_r(c, self.alphabet, int(m))
 
             # Final steps
             B = A
@@ -384,23 +342,83 @@ class FF3Cipher:
 
         return A + B
 
-DIGITS = string.digits + string.ascii_lowercase 
-LEN_DIGITS = len(DIGITS)
+def calculate_p(i, alphabet, W, B):
+    # P is always 16 bytes
+    P = bytearray(BLOCK_SIZE)
 
-def base_conv_r(n, base=2, length=0):    
+    # Calculate P by XORing W, i into the first 4 bytes of P
+    # i only requires 1 byte, rest are 0 padding bytes
+    # Anything XOR 0 is itself, so only need to XOR the last byte
+
+    P[0] = W[0]
+    P[1] = W[1]
+    P[2] = W[2]
+    P[3] = W[3] ^ int(i)
+
+    # The remaining 12 bytes of P are for rev(B) with padding
+
+    BBytes = decode_int_r(B, alphabet).to_bytes(12, "big")
+    # logging.debug(f"B: {B} BBytes: {BBytes.hex()}")
+
+    P[BLOCK_SIZE - len(BBytes):] = BBytes
+    return P
+
+def calculate_tweak64_ff3_1(tweak56):
+    tweak64 = bytearray(8)
+    tweak64[0] = tweak56[0]
+    tweak64[1] = tweak56[1]
+    tweak64[2] = tweak56[2]
+    tweak64[3] = (tweak56[3] & 0xF0)
+    tweak64[4] = tweak56[4]
+    tweak64[5] = tweak56[5]
+    tweak64[6] = tweak56[6]
+    tweak64[7] = ((tweak56[3] & 0x0F) << 4)
+    return tweak64
+
+def encode_int_r(n, alphabet, length=0):
     """
-    Return a string representation of a number in the given base system for 2..36
+    Return a string representation of a number in the given base system for 2..62
 
     The string is left in a reversed order expected by the calling cryptographic function
+
+    examples:
+       encode_int_r(10, hexdigits)
+        'A'
     """
+    base = len(alphabet)
+    if (base > FF3Cipher.RADIX_MAX):
+        raise ValueError(f"Base {base} is outside range of supported radix 2..{FF3Cipher.RADIX_MAX}")
 
     x = ''
     while n >= base:
         n, b = divmod(n, base)
-        x += DIGITS[b]
-    x += DIGITS[n]
+        x += alphabet[b]
+    x += alphabet[n]
 
-    if (len(x) < length):
-        x=x.ljust(length,'0')
+    if len(x) < length:
+        x = x.ljust(length, alphabet[0])
 
     return x
+
+
+def decode_int_r(astring, alphabet):
+    """Decode a Base X encoded string into the number
+
+    Arguments:
+    - `astring`: The encoded string
+    - `alphabet`: The alphabet to use for decoding
+    """
+    strlen = len(astring)
+    base = len(alphabet)
+    num = 0
+
+    idx = 0
+    try:
+        for char in reversed(astring):
+            power = (strlen - (idx + 1))
+            num += alphabet.index(char) * (base ** power)
+            idx += 1
+    except ValueError:
+        raise ValueError(f'char {char} not found in alphabet {alphabet}')
+
+    return num
